@@ -467,7 +467,9 @@ class BasePanel:
                 except AttributeError: pass
                 else: src = component.sigout
                 try: src
-                except NameError: continue
+                except NameError:
+                    console.assign(child)
+                    continue
                 #
                 try: child.sig_in
                 except AttributeError: pass
@@ -476,7 +478,9 @@ class BasePanel:
                 except AttributeError: pass
                 else: dst = child.sigin
                 try: dst
-                except NameError: continue
+                except NameError:
+                    console.assign(child)
+                    continue
                 src.link(dst)
             console.assign(child)
     def assign(self,component):
